/*
 * asteriskaudio.cxx
 *
 * Sound driver implementation for ASTERISK PBX.
 * OpenH323 Wrapper Library.
 *
 * Copyright (c) 2002,2003 InAccess Networks
 * Michalis Manousos <manousos@inaccessnetworks.com>
 * Dimitris Economou <decon@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: asteriskaudio.cxx,v 1.17 2003/03/05 12:37:36 manousos Exp $
 *
 */

#include <ptlib.h>
#include <sys/time.h>
#include <rtp.h>

#include "oh323.h"
#include "asteriskaudio.hxx"


/***********************************************************
 * Class PAsteriskAudioDelay
 **********************************************************/

PAsteriskAudioDelay::PAsteriskAudioDelay()
{
	WRAPTRACE(5, "Object initialized.");

	firstTime = TRUE;
	error = 0;
}
    
PAsteriskAudioDelay::~PAsteriskAudioDelay()
{
	WRAPTRACE(5, "Object deleted.");
}
    
void PAsteriskAudioDelay::Restart()
{
	firstTime = TRUE;
}
  
BOOL PAsteriskAudioDelay::Delay(int frameTime)
{
	if (firstTime) {
		firstTime = FALSE;
		previousTime = PTime();
		return TRUE;
	}

	error += frameTime;

	PTime now;
	PTimeInterval delay = now - previousTime;
	error -= (int)delay.GetMilliSeconds();
	previousTime = now;

	if (error > 0)
#ifdef P_LINUX
		usleep(error * 1000);
#else
		PThread::Current()->Sleep(error);
#endif

	return error <= -frameTime;
}


/***********************************************************
 * Class PAsteriskSoundChannel
 **********************************************************/

PAsteriskSoundChannel::PAsteriskSoundChannel()
{
	WRAPTRACE(5, "Object initialized.");
	mutex_flag = 0;
	lastWritten = 0;

	Construct();
}

void PAsteriskSoundChannel::Construct()
{
}

PAsteriskSoundChannel::~PAsteriskSoundChannel()
{
	Close();
	WRAPTRACE(5, "Object deleted.");
}

PStringArray PAsteriskSoundChannel::GetDeviceNames(Directions dir)
{
	PStringArray	recordArray, playArray;
	PStringArray	error;
	int				i;
	char			buf[MAX_FIFO_NAME];

	error[0] = "";

	/* Fill the recordArray */
	for (i=0; i<AST_OH323_MAX_CALLS; i++) {
		memset(buf, 0, MAX_FIFO_NAME);
		snprintf(buf, MAX_FIFO_NAME-1, "%sin%d", AST_OH323_FIFO_PREFIX, i);
		recordArray[i] = buf;
	}

	/* Fill the playArray */
	for (i=0; i<AST_OH323_MAX_CALLS; i++) {
		memset(buf, 0, MAX_FIFO_NAME);
		snprintf(buf, MAX_FIFO_NAME-1, "%sout%d", AST_OH323_FIFO_PREFIX, i);
		playArray[i] = buf;
	}

	if (dir == Recorder) {
		return recordArray;
	} else if (dir == Player) {
		return playArray;
	} else {
		PTRACE(1, "The direction entered is invalid.\n");
		return error;
	}
}

PString PAsteriskSoundChannel::GetDefaultDevice(Directions dir)
{
	PString		res = AST_OH323_FIFO_PREFIX;

	if (dir == Recorder) {
		res += "in0";
		return res;
	} else if (dir == Player) {
		res += "out0";
		return res;
	} else {
		PTRACE(1, "The direction entered is invalid.\n");
		return "";
	}
}

BOOL PAsteriskSoundChannel::Open(const PString & device,
								 int deviceFd,
								 Directions dir,
								 unsigned numChannels,
								 unsigned mediaFmt,
								 unsigned frameTm)
{
	Close();

	/***********
	if (!ConvertOSError(os_handle = ::open(device, dir == Player ? O_WRONLY : O_RDONLY | O_NONBLOCK)))
		return FALSE;
	**********/
	if (deviceFd < 0)
		return FALSE;
	os_handle = deviceFd;
	mediaFormat = mediaFmt;
	frameTime = frameTm;
	WRAPTRACE(5, "os_handle " << os_handle << ", mediaFormat " << mediaFormat
				<< ", frameTime " << frameTime);
	return TRUE;
	//return SetFormat(numChannels, sampleRate, bitsPerSample);
}

BOOL PAsteriskSoundChannel::Close()
{
	// XXX Do not close it from in here, do it from the channel driver.
	//os_handle = -1;
	//return TRUE;
	if (PChannel::Close() != TRUE) {
		// WRAPTRACE(10, "Failed - " << GetErrorText());
		return FALSE;
	} else
		return TRUE;
}

BOOL PAsteriskSoundChannel::SetFormat(unsigned numChannels,
							  unsigned sampleRate,
							  unsigned bitsPerSample)
{
	Abort();
	PAssert(numChannels >= 1 && numChannels <= 2, PInvalidParameter);
	PAssert(bitsPerSample == 8 || bitsPerSample == 16, PInvalidParameter);
	return TRUE;
}

BOOL PAsteriskSoundChannel::SetBuffers(PINDEX size, PINDEX count)
{
	Abort();
	PAssert(size > 0 && count > 0 && count < 65536, PInvalidParameter);
	bufSize = size;
	bufCount = count;

	return TRUE;
}

BOOL PAsteriskSoundChannel::GetBuffers(PINDEX & size, PINDEX & count)
{
	size = bufSize;
	count = bufCount;

	return TRUE;
}
#if 0
BOOL PAsteriskSoundChannel::Write(const void * buffer, PINDEX length)
{
	//WRAPTRACE(10, "In Write... " << length);
	if (PChannel::Write(buffer, length) != TRUE) {
		WRAPTRACE(3, "Failed - " << GetErrorText());
		return FALSE;
	} else {
		// Do a sleep to simulate the amount of time the hardware would take
		// to write out this data.
		switch (mediaFormat) {
			//case RTP_DataFrame::L16_Mono:
				// 1 msec/frame
				// 1 frame=16 bytes
			//	astWriteDelay.Delay(length/16);
			//	break;
			case RTP_DataFrame::PCMU:
			case RTP_DataFrame::PCMA:
				//astWriteDelay.Delay(length/16);
				astWriteDelay.Delay(length/8);
				break;
			case RTP_DataFrame::G729:
				// 10 msec/frame
				astWriteDelay.Delay(10);
				break;
			case RTP_DataFrame::GSM:
				// 20 msec/frame
				astWriteDelay.Delay(20);
				break;
			case RTP_DataFrame::G7231:
				// 30 msec/frame
				astWriteDelay.Delay(30);
				break;
			default:
				WRAPTRACE(3, "Unknown media format " << mediaFormat);
				break;
		}

		return TRUE;
	}
}
#endif

static unsigned const G7231FrameSize[4] = { 24, 20, 4, 1};
BOOL PAsteriskSoundChannel::Write(const void * buffer, PINDEX length)
{
	int written, size, num;
	char *buf;

	// Do a sleep to simulate the amount of time the hardware would take
	// to write out this data.
	lastWritten = 0;
	switch (mediaFormat) {
		//case RTP_DataFrame::L16_Mono:
			// 1 msec/frame
			// 1 frame=16 bytes
		//	astWriteDelay.Delay(length/16);
		//	break;
		case RTP_DataFrame::PCMU:
		case RTP_DataFrame::PCMA:
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(3, "Write failed (G.711) - " << GetErrorText());
				return FALSE;
			} else {
				// 1 msec/frame
				num = length/8;
				astWriteDelay.Delay(num * frameTime);
				lastWritten = length;
			}
			break;
		case RTP_DataFrame::G729:
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(3, "Write failed (G.729) - " << GetErrorText());
				return FALSE;
			} else {
				// 10 msec/frame
				astWriteDelay.Delay(frameTime);
				lastWritten = length;
			}
			break;
		case RTP_DataFrame::GSM:
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(3, "Write failed (GSM) - " << GetErrorText());
				return FALSE;
			} else {
				// 20 msec/frame
				num = length/33;
				astWriteDelay.Delay(num * frameTime);
				lastWritten = length;
			}
			break;
		case RTP_DataFrame::G7231:
#if 0
			//WRAPTRACE(10, "Length is " << length);
			if (PChannel::Write(buffer, length) != TRUE) {
				WRAPTRACE(3, "Write failed (G.723.1) - " << GetErrorText());
				return FALSE;
			} else {
				// 30 msec/frame
				astWriteDelay.Delay(30);
				lastWritten = length;
			}
			break;
#endif
			buf = (char*)buffer;
			size = length;
			//WRAPTRACE(10, "Length is " << length);
			while (size > 0) {
				written = G7231FrameSize[buf[length-size]&3];
				//WRAPTRACE(10, "Writing " << written);
				if (PChannel::Write((const void *)&buf[length-size], written) != TRUE) {
					WRAPTRACE(3, "Write failed (G.723.1) - " << GetErrorText());
					return FALSE;
				} else {
					// 30 msec/frame
					astWriteDelay.Delay(frameTime);
				}
				size -= written;
				lastWritten += written;
			}
			break;
		default:
			WRAPTRACE(3, "Unknown media format " << mediaFormat);
			break;
	}
	return TRUE;
}

PINDEX PAsteriskSoundChannel::GetLastWriteCount() const
{
	WRAPTRACE(10, "Returning " << lastWritten);
	return(lastWritten < 0 ? 0 : lastWritten);
}

BOOL PAsteriskSoundChannel::PlaySound(const PSound & sound, BOOL wait)
{
	Abort();

	if (!Write((const BYTE *)sound, sound.GetSize()))
		return FALSE;

	if (wait)
		return WaitForPlayCompletion();

	return TRUE;
}

BOOL PAsteriskSoundChannel::PlayFile(const PFilePath & filename, BOOL wait)
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::HasPlayCompleted()
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::WaitForPlayCompletion()
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::Read(void * buffer, PINDEX length)
{
	//WRAPTRACE(10, "In Read...");
	if (mutex_flag != 0)
		WRAPTRACE(3, "Previous read has not exited!\n");

	mutex_flag = 1;

	//WRAPTRACE(10, "length=" << length);
	if (PChannel::Read(buffer, length) != TRUE) {
		WRAPTRACE(3, "Failed - " << GetErrorText());
		mutex_flag = 0;
		return FALSE;
	} else {
		// This sleep simulates the time taken to read from real hardware
		//astReadDelay.Delay(length/16);
		mutex_flag = 0;
		return TRUE;
	}
}

BOOL PAsteriskSoundChannel::RecordSound(PSound & sound)
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::RecordFile(const PFilePath & filename)
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::StartRecording()
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::IsRecordBufferFull()
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::AreAllRecordBuffersFull()
{
	return TRUE;
}

BOOL PAsteriskSoundChannel::WaitForRecordBufferFull()
{
	if (os_handle < 0) {
		return FALSE;
	}

	return PXSetIOBlock(PXReadBlock, readTimeout);
}

BOOL PAsteriskSoundChannel::WaitForAllRecordBuffersFull()
{
	return FALSE;
}

BOOL PAsteriskSoundChannel::Abort()
{
	return TRUE;
}

// End of file /////////////////////////////////////////////////////////////
