/*
 * wrapconnection.cxx
 *
 * OpenH323 Wrapper Library
 *
 * Copyright (c) 2002,2003 InAccess Networks
 * Michalis Manousos <manousos@inaccessnetworks.com>
 * Dimitris Economou <decon@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: wrapconnection.cxx,v 1.5 2003/03/05 12:37:36 manousos Exp $
 *
 */

/************************************************************************/

#include <ptlib.h>
#include <h323.h>
#include <h323pdu.h>
#include <mediafmt.h>
#include <lid.h>

#define cplusplus
#include "wrapper.hxx"
#include "wrapendpoint.hxx"
#include "wrapconnection.hxx"
#include "asteriskaudio.hxx"
#include "version.h"

using namespace std;

// External stuff used inside the wrapper library
extern int		channelsOpen;
extern PMutex	*channelMutex;
extern BOOL		Con_noFastStart;
extern BOOL		Con_noH245Tunnelling;
extern BOOL		Con_noH245InSetup;
//extern BOOL		Con_noInBandDTMF;
extern WORD		Con_jitterMin;
extern WORD		Con_jitterMax;
extern BOOL		Con_noSilenceSuppression;
extern start_logchan_cb		on_start_logical_channel; 
extern clear_con_cb			on_connection_cleared;
extern alert_con_cb			on_connection_alert;
extern h323_exception_cb	on_h323_exception;
extern init_con_cb			on_connection_init;

/********************************************
 * Class WrapH323Connection
 ********************************************/

WrapH323Connection::WrapH323Connection(WrapH323EndPoint & ep,
				   unsigned callReference)
: H323Connection(ep, callReference)
{
	// For outgoing calls: 
	//     Remove all capabilities and add only the one requested
	//     Set the caller ID
	if (ep.MakingCall) {
		if (ep.MakeCallCapability != NULL) {
			localCapabilities.RemoveAll();
			localCapabilities.SetCapability(0, 0, ep.MakeCallCapability);
			WRAPTRACE(10, "Outgoing call with capability " << 
						(ep.MakeCallCapability)->GetFormatName());
		}
#if 0
		if (strlen(ep.MakeCallCID) != 0) {
			PString  cid = PString(ep.MakeCallCID);
			SetLocalPartyName(cid);
			WRAPTRACE(10, "Caller ID on outgoing call " << cid);
			WRAPTRACE(10, "LocalPartyName " << localPartyName);
			WRAPTRACE(10, "DestExtraCallInfo " << destExtraCallInfo);
		}
#endif
	}

	WRAPTRACE(6, "WrapH323Connection created.");
	return;
}

WrapH323Connection::~WrapH323Connection()
{
	WRAPTRACE(6, "WrapH323Connection deleted.");
	return;
}

H323Connection::AnswerCallResponse
WrapH323Connection::OnAnswerCall(const PString & caller,
					const H323SignalPDU & /*setupPDU*/,
					H323SignalPDU & /*connectPDU*/)
{
	call_details_t	cdet;		// Call details structure is passed to the PBX app
	int				res;

	// All the incoming calls receive an ALERTING PDU.
	// The call can be answered later with "AnsweringCall()" function.
	WRAPTRACE(6, "User " << caller << " is calling us...");

	// Fill the structure
	cdet.call_reference = GetCallReference();
	cdet.call_token = (const char *)GetCallToken();
	cdet.call_source_alias = (const char *)sourceAliases;
	cdet.call_dest_alias = (const char *)destAliases;
	cdet.call_source_e164 = (const char *)sourceE164;
	cdet.call_dest_e164 = (const char *)destE164;
	// DEBUG START
	WRAPTRACE(6, "Call reference: " << cdet.call_reference);
	WRAPTRACE(6, "Call token: " << cdet.call_token);
	WRAPTRACE(6, "Call source alias: "<< cdet.call_source_alias<<"("<<strlen(cdet.call_source_alias)<<")");
	WRAPTRACE(6, "Call dest alias: "<< cdet.call_dest_alias<<"("<<strlen(cdet.call_dest_alias)<<")");
	WRAPTRACE(6, "Call source e164: "<< cdet.call_source_e164<<"("<<strlen(cdet.call_source_e164)<<")");
	WRAPTRACE(6, "Call dest e164: "<< cdet.call_dest_e164<<"("<<strlen(cdet.call_dest_e164)<<")");
	WRAPTRACE(6, "Remote Party number: "<<GetRemotePartyNumber());
	WRAPTRACE(6, "Remote Party name: "<<GetRemotePartyName());
	WRAPTRACE(6, "Remote Party address: "<<GetRemotePartyAddress());
	// DEBUG END

	// Invoke the application callback
	res = on_connection_init(cdet);
	if (res < 0) {
		WRAPTRACE(6, "Failed to initialize incoming H.323 call. Dropping it.");
		return H323Connection::AnswerCallDenied;
	}

	//return H323Connection::AnswerCallNow;
	return H323Connection::AnswerCallPending;
}

BOOL 
WrapH323Connection::OnAlerting(const H323SignalPDU & /*alertingPDU*/,
		const PString & username)
{
	call_details_t cdet;

	WRAPTRACE(6, "Ringing phone for \"" << username << "\" ...");

	// Invoke the PBX application registered callback
	cdet.call_reference = GetCallReference();
	cdet.call_token = (const char *)GetCallToken();
	on_connection_alert(cdet);

	return TRUE;
}

BOOL 
WrapH323Connection::OnReceivedSignalSetup(const H323SignalPDU & setupPDU)
{
	WRAPTRACE(6, "Received SETUP message...");
	sourceAliases = setupPDU.GetSourceAliases();
	destAliases = setupPDU.GetDestinationAlias();
	sourceE164 = "";
	setupPDU.GetSourceE164(sourceE164);
	destE164 = "";
	setupPDU.GetDestinationE164(destE164);

	if (H323Connection::OnReceivedSignalSetup(setupPDU)==FALSE){
		WRAPTRACE(6, "Returning FALSE.");
		return FALSE;
	}
	return TRUE;
}

BOOL 
WrapH323Connection::OnSendSignalSetup(H323SignalPDU & setupPDU)
{
	WRAPTRACE(6, "Sending SETUP message...");
	sourceAliases = setupPDU.GetSourceAliases();
	destAliases = setupPDU.GetDestinationAlias();
	sourceE164 = "";
	setupPDU.GetSourceE164(sourceE164);
	destE164 = "";
	setupPDU.GetDestinationE164(destE164);

	return H323Connection::OnSendSignalSetup(setupPDU);
}

void
WrapH323Connection::OnReceivedReleaseComplete(const H323SignalPDU & pdu)
{
	WRAPTRACE(6, "Received RELEASE COMPLETE message...");

	return H323Connection::OnReceivedReleaseComplete(pdu);
}

BOOL
WrapH323Connection::OnClosingLogicalChannel(H323Channel & channel)
{
	WRAPTRACE(6, "Closing logical channel...");
	if (channel.GetCodec() != NULL) {
		(channel.GetCodec())->CloseRawDataChannel();
	}

	return H323Connection::OnClosingLogicalChannel(channel);
}

/***************
H323Channel *
WrapH323Connection::CreateLogicalChannel(const H245_OpenLogicalChannel & open,
										 BOOL startingFast,
										 unsigned & errorCode)
{
	WRAPTRACE(6, "Creating logical channel...");
	pChannel = H323Connection::CreateLogicalChannel(open, startingFast, errorCode);
	return pChannel;
}
**************/
// End of file //////////////////////////////////////////////////////////////
