/*
 * wrapendpoint.cxx
 *
 * OpenH323 Wrapper Library
 *
 * Copyright (c) 2002,2003 InAccess Networks
 * Michalis Manousos <manousos@inaccessnetworks.com>
 * Dimitris Economou <decon@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: wrapendpoint.cxx,v 1.10 2003/03/29 13:16:40 manousos Exp $
 *
 */

/************************************************************************/

#include <ptlib.h>
#include <h323.h>
#include <h323pdu.h>
#include <mediafmt.h>
#include <lid.h>

#define cplusplus
#include "wrapper.hxx"
#include "wrapendpoint.hxx"
#include "wrapconnection.hxx"
#include "asteriskaudio.hxx"
#include "version.h"

using namespace std;

// External stuff used inside the wrapper library
extern int		channelsOpen;
extern PMutex	*channelMutex;
extern BOOL		Con_noFastStart;
extern BOOL		Con_noH245Tunnelling;
extern BOOL		Con_noH245inSetup;
//extern BOOL		Con_noInBandDTMF;
extern WORD		Con_jitterMin;
extern WORD		Con_jitterMax;
extern BOOL		Con_noSilenceSuppression;
extern start_logchan_cb		on_start_logical_channel; 
extern clear_con_cb			on_connection_cleared;
extern alert_con_cb			on_connection_alert;
extern h323_exception_cb	on_h323_exception;
extern init_con_cb			on_connection_init;

/*******************************************************
 * Class WrapH323EndPoint 
 *******************************************************/

/**
 * Set terminal type to GATEWAY and add all
 * GW prefixes.
 */
WrapH323EndPoint::WrapH323EndPoint(char **gwprefix_tab, int prefix_num)
	: H323EndPoint()
{
	// Extensive tracing
	//PTrace::Initialise(5, "/tmp/lala");

	terminalType = e_GatewayOnly;
	MakingCall = FALSE;

	if (prefix_num > 0) {
		GWPrefix = new PStringArray(prefix_num, gwprefix_tab);
		if (GWPrefix == NULL)
			WRAPTRACE(3, "Failed to create GW prefixes list.");
		/**** DEBUG
		for (int i=0; i<GWPrefix->GetSize(); i++)
			cout << "Prefix " << i << " is " << (*GWPrefix)[i] << endl;
		*****/
	} else
		GWPrefix = NULL;

	frames_L16 = 30;
	frames_PCMU = 30;
	frames_PCMA = 30;
	frames_G728 = 1;
	frames_G729 = 1;
	frames_G7231 = 1;
	frames_GSM = 4;

	WRAPTRACE(5, "Compile-time libraries OpenH323 v" << OPENH323VERSION
				<< ", PWlib v" << PWLIBVERSION);
}

void
WrapH323EndPoint::SetEndpointTypeInfo(H225_EndpointType & info) const
{
	H323EndPoint::SetEndpointTypeInfo(info);

	info.m_gateway.IncludeOptionalField(H225_GatewayInfo::e_protocol);
	info.m_gateway.m_protocol.SetSize(1);
	H225_SupportedProtocols &protocol=info.m_gateway.m_protocol[0];
	protocol.SetTag(H225_SupportedProtocols::e_voice);

	if (GWPrefix != NULL) {
		PINDEX as=GWPrefix->GetSize();
		((H225_VoiceCaps &)protocol).m_supportedPrefixes.SetSize(as);
		for (PINDEX p=0; p<as; p++)
			H323SetAliasAddress((*GWPrefix)[p],
					((H225_VoiceCaps &)protocol).m_supportedPrefixes[p].m_prefix);
	}
}

/**
 * Return the number of frames per packet for the specified format,
 * -1 on failure.
 */
int
WrapH323EndPoint::GetFrames(const OpalMediaFormat format)
{
	int	res;

	res = -1;
	switch (format.GetPayloadType()) {
		case RTP_DataFrame::L16_Mono:
			res = frames_L16;
			break;
		case RTP_DataFrame::PCMU:
			res = frames_PCMU;
			break;
		case RTP_DataFrame::PCMA:
			res = frames_PCMA;
			break;
		case RTP_DataFrame::G728:
			res = frames_G728;
			break;
		case RTP_DataFrame::G729:
			res = frames_G729;
			break;
		case RTP_DataFrame::G7231:
			res = frames_G7231;
			break;
		case RTP_DataFrame::GSM:
			res = frames_GSM;
			break;
		default:
			break;
	}
	//WRAPTRACE(5, "Returning " << res);
	return(res);
}

/**
 * Store the number of frames per packet for the specified format,
 * -1 on failure.
 */
void
WrapH323EndPoint::SetFrames(const OpalMediaFormat format, int frames)
{
	//WRAPTRACE(5, "Setting " << frames);
	switch (format.GetPayloadType()) {
		case RTP_DataFrame::L16_Mono:
			frames_L16 = frames;
			break;
		case RTP_DataFrame::PCMU:
			frames_PCMU = frames;
			break;
		case RTP_DataFrame::PCMA:
			frames_PCMA = frames;
			break;
		case RTP_DataFrame::G728:
			frames_G728 = frames;
			break;
		case RTP_DataFrame::G729:
			frames_G729 = frames;
			break;
		case RTP_DataFrame::G7231:
			frames_G7231 = frames;
			break;
		case RTP_DataFrame::GSM:
			frames_GSM = frames;
			break;
		default:
			break;
	}
}

/**
 * Convert a OPAL media format to a OpenH323 wrapper library value
 */
cap_type_t
WrapH323EndPoint::GetCodecFromFormat(OpalMediaFormat &format)
{
	cap_type_t	res;

	res = CAP_UNDEFINED;
	switch (format.GetPayloadType()) {
		case RTP_DataFrame::L16_Mono:
			res = LINEAR16;
			break;
		case RTP_DataFrame::PCMU:
			res = G711U;
			break;
		case RTP_DataFrame::PCMA:
			res = G711A;
			break;
		case RTP_DataFrame::G728:
			res = G728;
			break;
		case RTP_DataFrame::G729:
			res = G729;
			break;
		case RTP_DataFrame::G7231:
			res = G7231;
			break;
		case RTP_DataFrame::GSM:
			res = GSM0610;
			break;
		default:
			break;
	}
	return(res);
}

/**
 * The fullAddress parameter is used directly in the MakeCall method so
 * the General form for the fullAddress argument is :
 * [alias@][transport$]host[:port]
 * default values:	alias = the same value as host.
 *					transport = ip.
 *					port = 1720.
 */
call_ret_val_t 
WrapH323EndPoint::MakeCall(const PString & dest, PString & token, 
					unsigned int *callReference, unsigned int port,
					H323Capability *h323Cap, char *CID)
{
	call_ret_val_t	retVal;
	PString			fullAddress;
	H323Connection	*pH323Con;

	/* Determine whether we are using a gatekeeper or not. */
	if (GetGatekeeper() != NULL) {
		fullAddress = dest;
		WRAPTRACE(5, "Making call to " << fullAddress << " using gatekeeper.");
	} else {
		fullAddress = dest; /* host */
		fullAddress += psprintf(":%i", port); /* host:port */
		/*fullAddress = dest.Trim() + '@' + fullAddress;*/ /* host@host:port */
		WRAPTRACE(5, "Making call to " << fullAddress << ".");
	}

	/* Initialize some call options */
	MakingCall = TRUE;
	MakeCallCapability = h323Cap;

	/* Set the CID, if available */
	if (CID != NULL)
		if (strlen(CID) != 0) {
			memset(MakeCallCID, 0, WRAPENDPOINT_CID_LEN);
			strncpy(MakeCallCID, CID, WRAPENDPOINT_CID_LEN-1);
			SetLocalUserName(PString(MakeCallCID));
		}
	/*if (!H323EndPoint::MakeCall(fullAddress, token)) {*/
	pH323Con = H323EndPoint::MakeCallLocked(fullAddress, token);
	if (pH323Con == NULL) {
		WRAPTRACE(5, "Error making call to \"" << fullAddress << '"');
		MakingCall = FALSE;
		MakeCallCapability = NULL;
		retVal = CALL_START_ER;
		return retVal;
	}
	/*pH323Con = FindConnectionWithLock(token);*/
	if (pH323Con != NULL) {
		*callReference = pH323Con->GetCallReference();
		pH323Con->Unlock();
		WRAPTRACE(5, GetLocalUserName() << " is calling host " << fullAddress);
		WRAPTRACE(6, "Call token is " << (const char *)token);
		WRAPTRACE(6, "Call reference is " << *callReference);
		retVal = CALL_START_OK;
	} else {
		WRAPTRACE(5, "Error making call to \"" << fullAddress << '"');
		retVal = CALL_START_ER;
	}

	MakingCall = FALSE;
	MakeCallCapability = NULL;
	return retVal;
}

BOOL
WrapH323EndPoint::AnswerCall(const PString & token)
{
	H323Connection						*pH323Con;
	H323Connection::AnswerCallResponse	resp;

	WRAPTRACE(6, "Request to answer call with token " << token);
	pH323Con = FindConnectionWithLock(token);
	if (pH323Con == NULL) {
		WRAPTRACE(5, "Could not find connection with token " << token);
		return FALSE;
	}
	resp = H323Connection::AnswerCallNow;
	pH323Con->AnsweringCall(resp);
	pH323Con->Unlock();
	WRAPTRACE(6, "Call with token " << token << " answered");
	return TRUE;
}
	
BOOL
WrapH323EndPoint::ClearCall(const PString & token)
{
	WRAPTRACE(6, "Request to clear call with token " << token);
	return H323EndPoint::ClearCall(token);
}

BOOL
WrapH323EndPoint::ClearCallSynchronous(const PString & token)
{
	WRAPTRACE(6, "Request to clear call with token " << token);
	return H323EndPoint::ClearCallSynchronous(token);
}

void
WrapH323EndPoint::SendUserInput(const PString &token, const PString &input)
{
	H323Connection::SendUserInputModes	mode;
	H323Connection						*pH323Con = NULL;

	pH323Con = FindConnectionWithLock(token);
	if (pH323Con != NULL) {
		pH323Con->SendUserInput(input);
		mode = pH323Con->GetRealSendUserInputMode();
		pH323Con->Unlock();
		WRAPTRACE(6, "Sent user input string (" << input << ") using mode " << mode);
	} else {
		WRAPTRACE(6, "No H323 connection with token " << token);
	}
}

void
WrapH323EndPoint::OnUserInputString(H323Connection & connection, 
			const PString & value)
{
	char	tone = value[0];

	WRAPTRACE(6, "Received user input string (" << value << ") from remote");

	// Invoke the callback function.
	on_h323_exception(connection.GetCallReference(), OH323EXC_USER_INPUT_TONE, &tone);
}

/**
 * This member function overrides the H323EndPoint::SetSoundChannelPlayDevice
 * in order to use our PAsteriskSoundChannel class.
 */
BOOL
WrapH323EndPoint::SetSoundChannelPlayDevice(const PString & name)
{
	if (PAsteriskSoundChannel::GetDeviceNames(PSoundChannel::Player).GetValuesIndex(name) == P_MAX_INDEX)
		return FALSE;
	soundChannelPlayDevice = name;
	return TRUE;
}

/**
 * This member function overrides the H323EndPoint::SetSoundChannelRecordDevice
 * in order to use our PAsteriskSoundChannel class.
 */
BOOL
WrapH323EndPoint::SetSoundChannelRecordDevice(const PString & name)
{
	if (PAsteriskSoundChannel::GetDeviceNames(PSoundChannel::Recorder).GetValuesIndex(name) == P_MAX_INDEX)
		return FALSE;
	soundChannelRecordDevice = name;
	return TRUE;
}

/** 
 * The OpenAudioChannel callback is invoked when the connection is
 * established, so the negotiations are over with success. The 
 * on_start_logical_channel callback which is registered from the 
 * pbx application is called and returns the name of the named FIFO
 * that the sound channel (PSoundChannel) will use to pass the voice
 * data 
 */
BOOL 
WrapH323EndPoint::OpenAudioChannel(H323Connection & connection,
		BOOL isEncoding,
		unsigned bufferSize,
		H323AudioCodec & codec)
{
	PString			deviceName, audio_dev_name;
	lchan_dir		dir;
	char			*dev_name;	/* name of the named FIFO */
	call_details_t	cdet;		/* Call details structure is passed to the PBX app. */
	int				fd;
	unsigned int	bufSize, bufNum;
	unsigned int	mediaType;
	unsigned int	frameTime;
	cap_type_t		cap_type;
	OpalMediaFormat	mediaFormat;
#if 0
	unsigned int	frameNum;
	H323Channel		*channel;

	channel = connection.FindChannel(RTP_Session::DefaultAudioSessionID, 
										connection.HadAnsweredCall());
	WRAPTRACE(6, " ***** Logical channel check");
	if (channel == NULL) {
		cout << " ***** Logical channel not found" << endl;
	} else {
		cout << " ***** Logical channel found" << endl; 
		if (isEncoding)
			frameNum = (channel->GetCapability()).GetTxFramesInPacket();
		else
			frameNum = (channel->GetCapability()).GetRxFramesInPacket();
		cout << " ***** Frames " << frameNum << endl; 
	}
	WRAPTRACE(6, " ***** Logical channel check end");
#endif

	dir = (isEncoding == true) ? RECORDER : PLAYER;

	/* Check direction */
	switch (dir) {
		case RECORDER:
			WRAPTRACE(6, "Direction => RECODER, Buffer => " << bufferSize);
			break;
		case PLAYER:
			WRAPTRACE(6, "Direction => PLAYER, Buffer => " << bufferSize);
			break;
		default:
			WRAPTRACE(6, "Direction => UNKNOWN");
			return FALSE;
	}

	/* Invoke the application callback and get the open
	   socket descriptor from ASTERISK. */
	cdet.call_reference = connection.GetCallReference();
	cdet.call_token = (const char *)connection.GetCallToken();
	cdet.call_source_alias = (const char *)((WrapH323Connection &)connection).sourceAliases;
	cdet.call_dest_alias = (const char *)((WrapH323Connection &)connection).destAliases;
	cdet.call_source_e164 = (const char *)((WrapH323Connection &)connection).sourceE164;
	cdet.call_dest_e164 = (const char *)((WrapH323Connection &)connection).destE164;

	mediaFormat = codec.GetMediaFormat();
	mediaType = mediaFormat.GetPayloadType();
	cap_type = GetCodecFromFormat(mediaFormat);
	if (cap_type == CAP_UNDEFINED) {
		WRAPTRACE(5, "Unknown/Unsupported media format" << mediaFormat);
		return FALSE;
	}
	if (codec.IsDescendant(H323_LIDCodec::Class())) {
		WRAPTRACE(5, "FrameSize " << mediaFormat.GetFrameSize() 
			<< ", FrameTime " << mediaFormat.GetFrameTime()
			<< ", TimeUnits " << mediaFormat.GetTimeUnits());
		frameTime = mediaFormat.GetFrameTime() / mediaFormat.GetTimeUnits();
		if (frameTime <= 0) {
			WRAPTRACE(3, "Cannot determine frame rate (" << frameTime << ")");
			return FALSE;
		}
		if (dir == RECORDER) {
			bufSize = GetFrames(mediaFormat) * mediaFormat.GetFrameSize();
			WRAPTRACE(5, "Frames " << GetFrames(mediaFormat));
		} else {
			if ((mediaType == RTP_DataFrame::PCMU) || (mediaType == RTP_DataFrame::PCMA)) {
				bufSize = bufferSize;
			} else {
				bufSize = mediaFormat.GetFrameSize();
				WRAPTRACE(5, "Frame 1");
			}
		}
		if (bufSize <= 0)
			bufSize = bufferSize;
		bufNum = 1;
		codec.SetSilenceDetectionMode(H323AudioCodec::NoSilenceDetection);
		WRAPTRACE(6, "LID Codec " << mediaFormat);
	} else {
		/* XXX Not used XXX */
		return FALSE;
		bufSize = bufferSize;
		bufNum = soundChannelBuffers;
		codec.SetSilenceDetectionMode(Con_noSilenceSuppression ?
								H323AudioCodec::NoSilenceDetection :
								H323AudioCodec::AdaptiveSilenceDetection);
		WRAPTRACE(6, "Audio Codec " << mediaFormat);
	}

	dev_name = NULL;
	dev_name = on_start_logical_channel(cdet, dir, bufSize, cap_type, &fd);
	if (dev_name == NULL) {
		WRAPTRACE(5, "Could not open a suitable sound channel with PBX.");
		return FALSE;
	}

	WRAPTRACE(6, "The sound channel is " << dev_name << "(fd=" << fd << ")");
	audio_dev_name = PString(dev_name);

	if (isEncoding) {
		SetSoundChannelRecordDevice(audio_dev_name);
		deviceName = GetSoundChannelRecordDevice();
	} else {
		SetSoundChannelPlayDevice(audio_dev_name);
		deviceName = GetSoundChannelPlayDevice();
	}
	WRAPTRACE(6, "The audio device name is " << deviceName);

	PAsteriskSoundChannel * soundChannel = new PAsteriskSoundChannel;

	/* Pass the socket descriptor for read or write according to isEncoding value */
	if (soundChannel->Open(deviceName, fd, 
				isEncoding ? PSoundChannel::Recorder : PSoundChannel::Player, 
				1, mediaType, frameTime)) {
		WRAPTRACE(6, "Opened sound channel \"" << deviceName
			<< "\" for " << (isEncoding ? "record" : "play")
			<< "ing using " << bufNum
			<< 'x' << bufSize << " byte buffers.");
		soundChannel->SetBuffers(bufSize, bufNum);

		/* Attaches the soundChannel on the socket with the selected codec */
		/* Delete sound channel on closing codec */
		return codec.AttachChannel(soundChannel, TRUE);
	}

	WRAPTRACE(5, "Could not open sound channel \"" << deviceName
	   << "\" for " << (isEncoding ? "record" : "play")
	   << "ing: " << soundChannel->GetErrorText());

	delete soundChannel;
	return FALSE;
}

/** 
 * This callback function is invoked once upon creation of each
 * channel for the two directions of an H323 session 
 */
BOOL 
WrapH323EndPoint::OnStartLogicalChannel(H323Connection & connection, H323Channel & channel)
{
	if (!H323EndPoint::OnStartLogicalChannel(connection, channel))
		return FALSE;

	WRAPTRACE(6, "Started logical channel [" << connection.GetCallReference() << "] : " 
		<< ((channel.GetDirection()==H323Channel::IsTransmitter)?"sending ":((channel.GetDirection()==H323Channel::IsReceiver)?"receiving ":" "))
		<< (const char *)(channel.GetCapability()).GetFormatName());

	// adjust the count of channels we have open
	if (channel.GetDirection()==H323Channel::IsTransmitter) {
		channelMutex->Wait();
		WRAPTRACE(6, "TxFrames = " << channel.GetCapability().GetTxFramesInPacket());
		channelsOpen++;
		WRAPTRACE(6, "channelsOpen = " << channelsOpen);
		channelMutex->Signal();
		return TRUE;
	}
	if (channel.GetDirection()==H323Channel::IsReceiver) {
		channelMutex->Wait();
		WRAPTRACE(6, "RxFrames = " << channel.GetCapability().GetRxFramesInPacket());
		channelsOpen++;
		WRAPTRACE(6, "channelsOpen = " << channelsOpen);
		channelMutex->Signal();
		return TRUE;
	}
	return FALSE;
}

void 
WrapH323EndPoint::OnClosedLogicalChannel(H323Connection & connection, 
				const H323Channel & channel)
{
	WRAPTRACE(6, "Closed logical channel [" << connection.GetCallReference() << "].");

	// adjust the count of channels we have open
	channelMutex->Wait();
	channelsOpen--;
	WRAPTRACE(6, "channelOpen = " << channelsOpen);
	channelMutex->Signal();
	H323EndPoint::OnClosedLogicalChannel(connection, channel);
}

/** 
 * The OnConnectionCleared callback function is called upon the dropping of an established
 * H323 connection. Among others it calls the on_connection_cleared function which is installed
 * previously by the PBX application and it is responsible for cleaning of any internal channel
 * related structuring 
 */
void 
WrapH323EndPoint::OnConnectionCleared(H323Connection & connection, const PString & clearedCallToken)
{
	PString			remoteName = '"' + connection.GetRemotePartyName() + '"';
	call_details_t	cdet;

	WRAPTRACE(6, "Closing connection [" << connection.GetCallReference() << "].");

	/* Invoke the PBX application registered callback */
	cdet.call_reference = connection.GetCallReference();
	cdet.call_token = (const char *)connection.GetCallToken();
	on_connection_cleared(cdet);

	switch (connection.GetCallEndReason()) {
		case H323Connection::EndedByCallForwarded :
			break;
		case H323Connection::EndedByRemoteUser :
			WRAPTRACE(5, remoteName << " has cleared the call");
			break;
		case H323Connection::EndedByCallerAbort :
			WRAPTRACE(5, remoteName << " has stopped calling");
			break;
		case H323Connection::EndedByRefusal :
			WRAPTRACE(5, remoteName << " did not accept your call");
			break;
		case H323Connection::EndedByRemoteBusy :
			WRAPTRACE(5, remoteName << " was busy");
			break;
		case H323Connection::EndedByRemoteCongestion :
			WRAPTRACE(5, "Congested link to " << remoteName);
			break;
		case H323Connection::EndedByNoAnswer :
			WRAPTRACE(5, remoteName << " did not answer your call");
			break;
		case H323Connection::EndedByTransportFail :
			WRAPTRACE(5, "Call with " << remoteName << " ended abnormally");
			break;
		case H323Connection::EndedByCapabilityExchange :
			WRAPTRACE(5, "Could not find common codec with " << remoteName);
			break;
		case H323Connection::EndedByNoAccept :
			WRAPTRACE(5, "Did not accept incoming call from " << remoteName);
			break;
		case H323Connection::EndedByAnswerDenied :
			WRAPTRACE(5, "Refused incoming call from " << remoteName);
			break;
		case H323Connection::EndedByNoUser :
			WRAPTRACE(5, "Gatekeeper could not find user " << remoteName);
			break;
		case H323Connection::EndedByNoBandwidth :
			WRAPTRACE(5, "Call to " << remoteName << " aborted, insufficient bandwidth.");
			break;
		case H323Connection::EndedByUnreachable :
			WRAPTRACE(5, remoteName << " could not be reached.");
			break;
		case H323Connection::EndedByHostOffline :
			WRAPTRACE(5, remoteName << " is not online.");
			break;
		case H323Connection::EndedByNoEndPoint :
			WRAPTRACE(5, "No phone running for " << remoteName);
			break;
		case H323Connection::EndedByConnectFail :
			WRAPTRACE(5, "Transport error calling " << remoteName);
			break;
		default :
			WRAPTRACE(5, "Call with " << remoteName << " completed");
	}

	PTime connectTime = connection.GetConnectionStartTime();
	if (connectTime.GetTimeInSeconds() != 0)
		WRAPTRACE(5, "Duration "
		 << setprecision(0) << setw(5)
		 << (PTime() - connection.GetConnectionStartTime()));
}

H323Connection *
WrapH323EndPoint::CreateConnection(unsigned callReference)
{
	return new WrapH323Connection(*this, callReference);
}

// End of file //////////////////////////////////////////////////////////////
