/*
 * wrapper.cxx
 *
 * OpenH323 Wrapper Library
 *
 * Copyright (c) 2002,2003 InAccess Networks
 * Michalis Manousos <manousos@inaccessnetworks.com>
 * Dimitris Economou <decon@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: wrapper.cxx,v 1.36 2003/03/17 17:16:31 manousos Exp $
 *
 */

/************************************************************************/
/* INCLUDE FILES ********************************************************/

#include <ptlib.h>
#include <h323.h>
#include <h323pdu.h>
#include <mediafmt.h>
#include <lid.h>

#include <list>
#include <string>
#include <algorithm>

#define cplusplus
#include "wrapper.hxx"
#include "wrapendpoint.hxx"
#include "wrapconnection.hxx"
#include "asteriskaudio.hxx"
#include "version.h"

using namespace std;

/************************************************************************/
/* Wrapper's internal variables and classes *****************************/

/* Counter for the number of connections */
int		channelsOpen;
PMutex	*channelMutex = NULL;

/* Options for connections creation */
BOOL	Con_noFastStart;
BOOL	Con_noH245Tunnelling;
BOOL	Con_noH245inSetup;
BOOL	Con_noInBandDTMF;
WORD	Con_jitterMin;
WORD	Con_jitterMax;
BOOL	Con_noSilenceSuppression;

/* Callback functions */
start_logchan_cb	on_start_logical_channel; 
clear_con_cb		on_connection_cleared;
alert_con_cb		on_connection_alert;
h323_exception_cb	on_h323_exception;
init_con_cb			on_connection_init;

/**
 * The wrapper library assumes that only one end point should exist.
 * The application cannot run twice the h323_end_point_create(). 
 */
WrapH323EndPoint *endPoint = NULL;

H323TransportUDP *rasChannel = NULL;

/** 
 * The capability class for holding the capability and its
 * description 
 */
class WrapCapability {

	public:
	H323Capability *capability;
	cap_type_t capabilityType;

	/* Constructor */
	WrapCapability(H323Capability *c, cap_type_t t)
	{
		capability = c;
		capabilityType = t;
	};
};

/**
 * A list with the H323 capabilities of the terminal 
 */
list<WrapCapability*> capabilityList; 

/** 
 * The internal class for holding the listener and its
 * description 
 */
class WrapListener {

	public:
	H323Listener *listener;
	lis_type_t listenerType;
    
	/* Constructor */
	WrapListener(H323Listener *l, lis_type_t t)
	{
		listener = l;
		listenerType = t;
	};
};

/**
 * A list with the H323 listeners of the end point 
 */
list<WrapListener*> listenerList; 


/**
 * The WrapThread is a necessary descendant of PProcess class so that the H323EndPoint 
 * object to be created from within that class. The reason is that the constructor
 * of the H323EndPoint class needs to find some information related to a hypothetically
 * PProcess class which is supposed to instantiates it. 
 */
class WrapProcess : public PProcess {

	PCLASSINFO(WrapProcess, PProcess);
    
	public:
	WrapProcess(char **gwprefix_tab, int gwprefix_num, int logLev, char *logFile)
	: PProcess("inAccess Networks (www.inaccessnetworks.com)", "OpenH323 Wrapper", 0, 0, AlphaCode, 0)
	{ 
		endPoint = NULL;
		rasChannel = NULL;
		prefixes = gwprefix_tab;
		prefix_num = gwprefix_num;
		log_lev = logLev;
		if ((logFile == NULL) || (logFile[0] == '\0'))
			log_file = NULL;
		else
			log_file = logFile;
		Resume(); 
	};

	~WrapProcess()
	{
		WRAPTRACE(5, "Going down.");
		if (endPoint != NULL) {
			delete endPoint;
			endPoint = NULL;
		}
		/*
		if (rasChannel != NULL) {
			delete rasChannel;
			rasChannel = NULL;
		}*/
		PTrace::SetLevel(0);
	}
    
	void 
	Main()
	{
		PTrace::Initialise(log_lev, log_file);
		endPoint = new WrapH323EndPoint(prefixes, prefix_num);
	};

	protected:
	char **prefixes;
	int prefix_num;
	int log_lev;
	char *log_file;
};
WrapProcess *localProcess = NULL;


/** 
 * This class handles the termination of a call.
 * Note that OpenH323 Library requires that the termination
 * of a call should be done inside a separate thread of execution.
 */
class ClearCallThread : public PThread {

	PCLASSINFO(ClearCallThread, PThread);

	public:
	ClearCallThread(const char *tc)
		: PThread(10000, PThread::AutoDeleteThread)
	{ 
		WRAPTRACE(5, "Object initialized.");
		token = tc;
		Resume(); 
	};

	~ClearCallThread()
	{
		WRAPTRACE(5, "Object deleted.");
		return;
	};
    
	void 
	Main()
	{
		//endPoint->ClearCall(token);
		if (endPoint->ClearCallSynchronous(token) == FALSE) {
			WRAPTRACE(5, "Failed to clear call with token " << token);
		} else {
			WRAPTRACE(5, "Call with token " << token << " cleared.");
		}
		return;
	};

	protected:
	PString	token;
};

/************************************************************************/
/* INTERNAL C-FUNCTIONS OF WRAPPER **************************************/

/**
 * Creates a H323Capability based on the given capability/codec type.
 * Returns a pointer to the capability on success, NULL on failure
 * or unsupported capability.
 */
H323Capability *
h323_capability_create(WrapH323EndPoint *ep, cap_type_t capability, int frame_num)
{
	H323Capability	*newCap;
	int frames;

	newCap = NULL;
	switch (capability) {
		case G711U:
		{
			// Use the built-in codec for G.711 u-law
			//H323_G711Capability *g711muCap = new H323_G711Capability(H323_G711Capability::muLaw);
			H323_LIDCapability *g711muCap = new H323_LIDCapability(OPAL_G711_ULAW_64K);
			if (frame_num > 0)
				ep->SetFrames(OpalG711uLaw, frame_num);
			frames = ep->GetFrames(OpalG711uLaw);
			if (frames <= 0) {
				delete g711muCap;
				break;
			}
			g711muCap->SetTxFramesInPacket(frames);
			newCap = g711muCap;
			break;
		}
		case G711A:
		{
			// Use the built-in codec for G.711 A-law
			//H323_G711Capability *g711aCap = new H323_G711Capability(H323_G711Capability::ALaw);
			H323_LIDCapability *g711aCap = new H323_LIDCapability(OPAL_G711_ALAW_64K);
			if (frame_num > 0)
				ep->SetFrames(OpalG711ALaw, frame_num);
			frames = ep->GetFrames(OpalG711ALaw);
			if (frames <= 0) {
				delete g711aCap;
				break;
			}
			g711aCap->SetTxFramesInPacket(frames);
			newCap = g711aCap;
			break;
		}
		case G7231:
		case G72316K3:
		{
			// G.723.1(6.3k) pass-through codec
			H323_LIDCapability *g72316k3Cap = new H323_LIDCapability(OPAL_G7231_6k3);
			if (frame_num > 0)
				ep->SetFrames(OpalG7231_6k3, frame_num);
			frames = ep->GetFrames(OpalG7231_6k3);
			if (frames <= 0) {
				delete g72316k3Cap;
				break;
			}
			g72316k3Cap->SetTxFramesInPacket(frames);
			newCap = g72316k3Cap;
			break;
		}
		case G72315K3:
		{
			// G.723.1(5.3k) pass-through codec
			H323_LIDCapability *g72315k3Cap = new H323_LIDCapability(OPAL_G7231_5k3);
			if (frame_num > 0)
				ep->SetFrames(OpalG7231_5k3, frame_num);
			frames = ep->GetFrames(OpalG7231_5k3);
			if (frames <= 0) {
				delete g72315k3Cap;
				break;
			}
			g72315k3Cap->SetTxFramesInPacket(frames);
			newCap = g72315k3Cap;
			break;
		}
		case G7231A6K3:
		{
			// G.723.1A(6.3k) pass-through codec
			H323_LIDCapability *g7231A6k3Cap = new H323_LIDCapability(OPAL_G7231A_6k3);
			if (frame_num > 0)
				ep->SetFrames(OpalG7231A_6k3, frame_num);
			frames = ep->GetFrames(OpalG7231A_6k3);
			if (frames <= 0) {
				delete g7231A6k3Cap;
				break;
			}
			g7231A6k3Cap->SetTxFramesInPacket(frames);
			newCap = g7231A6k3Cap;
			break;
		}
		case G7231A5K3:
		{
			// G.723.1A(5.3k) pass-through codec
			H323_LIDCapability *g7231A5k3Cap = new H323_LIDCapability(OPAL_G7231A_5k3);
			if (frame_num > 0)
				ep->SetFrames(OpalG7231A_5k3, frame_num);
			frames = ep->GetFrames(OpalG7231A_5k3);
			if (frames <= 0) {
				delete g7231A5k3Cap;
				break;
			}
			g7231A5k3Cap->SetTxFramesInPacket(frames);
			newCap = g7231A5k3Cap;
			break;
		}
		case G728:
		{
			// G.728 pass-through codec
			H323_LIDCapability *g728Cap = new H323_LIDCapability(OPAL_G728);
			if (frame_num > 0)
				ep->SetFrames(OpalG728, frame_num);
			frames = ep->GetFrames(OpalG728);
			if (frames <= 0) {
				delete g728Cap;
				break;
			}
			g728Cap->SetTxFramesInPacket(frames);
			newCap = g728Cap;
			break;
		}
		case G729:
		{
			// G.729 pass-through codec
			H323_LIDCapability *g729Cap = new H323_LIDCapability(OPAL_G729);
			if (frame_num > 0)
				ep->SetFrames(OpalG729, frame_num);
			frames = ep->GetFrames(OpalG729);
			if (frames <= 0) {
				delete g729Cap;
				break;
			}
			g729Cap->SetTxFramesInPacket(frames);
			newCap = g729Cap;
			break;
		}
		case G729A:
		{
			// G.729A pass-through codec
			H323_LIDCapability *g729ACap = new H323_LIDCapability(OPAL_G729A);
			if (frame_num > 0)
				ep->SetFrames(OpalG729A, frame_num);
			frames = ep->GetFrames(OpalG729A);
			if (frames <= 0) {
				delete g729ACap;
				break;
			}
			g729ACap->SetTxFramesInPacket(frames);
			newCap = g729ACap;
			break;
		}
		case G729B:
		{
			// G.729B pass-through codec
			H323_LIDCapability *g729BCap = new H323_LIDCapability(OPAL_G729B);
			if (frame_num > 0)
				ep->SetFrames(OpalG729B, frame_num);
			frames = ep->GetFrames(OpalG729B);
			if (frames <= 0) {
				delete g729BCap;
				break;
			}
			g729BCap->SetTxFramesInPacket(frames);
			newCap = g729BCap;
			break;
		}
		case G729AB:
		{
			// G.729AB pass-through codec
			H323_LIDCapability *g729ABCap = new H323_LIDCapability(OPAL_G729AB);
			if (frame_num > 0)
				ep->SetFrames(OpalG729AB, frame_num);
			frames = ep->GetFrames(OpalG729AB);
			if (frames <= 0) {
				delete g729ABCap;
				break;
			}
			g729ABCap->SetTxFramesInPacket(frames);
			newCap = g729ABCap;
			break;
		}
		case GSM0610:
		{
			// GSM 06.10 pass-through codec
			H323_LIDCapability *gsmCap = new H323_LIDCapability(OPAL_GSM0610);
			if (frame_num > 0)
				ep->SetFrames(OpalGSM0610, frame_num);
			frames = ep->GetFrames(OpalGSM0610);
			if (frames <= 0) {
				delete gsmCap;
				break;
			}
			gsmCap->SetTxFramesInPacket(frames);
			newCap = gsmCap;
			break;
		}
		case MSGSM:
		{
#if 0
			// Use the built-in codec for Microsoft GSM
			MicrosoftGSMAudioCapability * msGsmCap = new MicrosoftGSMAudioCapability;
			msGsmCap->SetTxFramesInPacket(4);
			newCap = msGsmCap;
#endif
			break;
		}
		case LPC10:
		{
#if 0
			// Use the built-in codec for LPC10
			if (ep == NULL)
				break;
			H323_LPC10Capability * lpcCap = new H323_LPC10Capability(*ep);
			newCap = lpcCap;
#endif
			break;
		}
		default:
			// Unknown codec!
			break;
	}
	return(newCap);
}

/************************************************************************/
/* IMPLEMENTATION OF WRAPPER API ****************************************/

/**
 * The extern "C" directive takes care for 
 * the ANSI-C representation of linkable symbols 
 */
extern "C" {

	/**
	 * Check whether the H.323 endpoint has already been created
	 */
	boolean_t 
	end_point_exist(void)
	{
		if (endPoint == NULL) {
			return NO;
		}
		return YES;
	}
 
	/**
	 * Create the H.323 endpoint
	 */
	void 
	h323_end_point_create(char **gwprefix_tab, int gwprefix_num, 
			int log_lev, char *log_file)
	{
		channelsOpen = 0;
		if (channelMutex == NULL)
			channelMutex = new PMutex();
		localProcess = new WrapProcess(gwprefix_tab, gwprefix_num, log_lev, log_file);
		localProcess->Main();
		WRAPTRACEAPI(5, "Endpoint created.");
	}

 	/**
	 * Destroy the H.323 endpoint
	 */
	void 
	h323_end_point_destroy(void)
	{
		WRAPTRACEAPI(5, "Destroying endpoint.");
		if (end_point_exist() == NO) {
			return;
		}

		/* Cleanup stuff */
		endPoint->ClearAllCalls();

		delete localProcess;
		localProcess = NULL;

		delete channelMutex;
		channelMutex = NULL;
	}

	/** 
	 * Install the callback functions
	 */
	int 
	h323_callback_register(start_logchan_cb sfunc, clear_con_cb cfunc,
						alert_con_cb alertfunc, h323_exception_cb exfunc,
						init_con_cb initfunc)
	{
		if (sfunc != NULL)
			on_start_logical_channel = sfunc;
		else {
			WRAPTRACEAPI(5, "on_start_logical_channel is NULL!");
			return(-1);
		}
		if (cfunc != NULL)
			on_connection_cleared = cfunc;
		else {
			WRAPTRACEAPI(5, "on_connection_cleared is NULL!");
			return(-1);
		}
		if (alertfunc != NULL)
			on_connection_alert = alertfunc;
		else {
			WRAPTRACEAPI(5, "on_connection_alert is NULL!");
			return(-1);
		}
		if (exfunc != NULL)
			on_h323_exception = exfunc;
		else {
			WRAPTRACEAPI(5, "on_h323_exception is NULL!");
			return(-1);
		}
		if (initfunc != NULL)
			on_connection_init = initfunc;
		else {
			WRAPTRACEAPI(5, "on_connection_init is NULL!");
			return(-1);
		}
		WRAPTRACEAPI(6, "Callback functions installed.");
		return(0);
	}

	/**
	 * The check_cap_type function is a logical predicate function
	 * for the find_if algorithm which is used in the h323_add_capability
	 * as well as in the h323_set_capability.
	 */
	static cap_type_t chkCap;

	static bool
	check_cap_type(WrapCapability *wc)
	{
		if (wc->capabilityType == chkCap) {
			return true;
		} else {
			return false;
		}
	}

	
    /** 
	 * Add the specified capability to the capabilities table of the H.323 endpoint 
	 * Note: AddCapability puts the codec into the list of codecs we can send
	 */
	cap_ret_val_t
	h323_add_capability(cap_type_t cap, int frames)
	{
		H323Capability					*h323Cap;
		WrapCapability 					*wrapCap;
		list<WrapCapability*>::iterator	cit;
		cap_ret_val_t					ret_val;

		if (end_point_exist() == NO) {
			return CAP_EP_ER;
		}

		ret_val = CAP_NSUP_ER;

		/* Check to see if a same type of capability exists in the capability list */
		chkCap = cap;
		cit = find_if(capabilityList.begin(), capabilityList.end(), check_cap_type);
 
		/* If the specified capability was found in the list, just return */
		if (cit != capabilityList.end()) {
			ret_val = CAP_EXIST;
			return ret_val;
		}
	
		/* Create and add the capability */
		h323Cap = h323_capability_create(endPoint, cap, frames);
		if (h323Cap != NULL) {
			endPoint->AddCapability(h323Cap);
			wrapCap = new WrapCapability(h323Cap, cap);
			capabilityList.insert(capabilityList.end(), wrapCap);
			WRAPTRACEAPI(10, "Added capability " << (wrapCap->capability)->GetFormatName());
			ret_val = CAP_INSERT_OK;
		} else {
			WRAPTRACEAPI(10, "Failed to add capability type " << cap);
			ret_val = CAP_NSUP_ER;
		}
		return ret_val;
	}

	/**
	 * Set the specified capability as the first capability of the H.323 endpoint
	 * Note: SetCapability puts the codec into the list of codecs we can send and receive
	 */
	cap_ret_val_t 
	h323_set_capability(cap_type_t cap, int frames)
	{
		H323Capability					*h323Cap;
		WrapCapability 					*wrapCap;
		list<WrapCapability*>::iterator	cit;
		cap_ret_val_t					ret_val;

		if (end_point_exist() == NO) {
			return CAP_EP_ER;
		}

		ret_val = CAP_NSUP_ER;

		/* Check to see if a same type of capability exists in the capability list */
		chkCap = cap;
		cit = find_if(capabilityList.begin(), capabilityList.end(), check_cap_type);
 
		/* If the specified capability was found in the list, just return */
		if (cit != capabilityList.end()) {
			ret_val = CAP_EXIST;
			return ret_val;
		}

		/* Create and add the capability */
		h323Cap = h323_capability_create(endPoint, cap, frames);
		if (h323Cap != NULL) {
			endPoint->SetCapability(0, 0, h323Cap);
			wrapCap = new WrapCapability(h323Cap, cap);
			capabilityList.insert(capabilityList.end(), wrapCap);
			WRAPTRACEAPI(10, "Inserted capability " << (wrapCap->capability)->GetFormatName());
			ret_val = CAP_INSERT_OK;
		} else {
			WRAPTRACEAPI(10, "Failed to insert capability type " << cap);
			ret_val = CAP_NSUP_ER;
		}
		return ret_val;
    }

	/**
	 * Set the mode for sending/receiving user-input (DTMF).
	 */
	cap_ret_val_t
	h323_set_senduimode(uimode_t cap)
	{
		if (end_point_exist() == NO) {
			return CAP_EP_ER;
		}

		/* Set the specified send user-input mode */
		switch (cap) {
			case UIMODE_Q931:
				endPoint->SetSendUserInputMode(H323Connection::SendUserInputAsQ931);
				break;
			case UIMODE_STRING:
				endPoint->SetSendUserInputMode(H323Connection::SendUserInputAsString);
				break;
			case UIMODE_TONE:
				endPoint->SetSendUserInputMode(H323Connection::SendUserInputAsTone);
				break;
			case UIMODE_RFC2833:
				endPoint->SetSendUserInputMode(H323Connection::SendUserInputAsInlineRFC2833);
				break;
			default:
				return CAP_NSUP_ER;
		}

		/* Add all user input capabilities (receiving user-input) */
		endPoint->AddAllUserInputCapabilities(0, P_MAX_INDEX);

		WRAPTRACEAPI(5, "User-input mode set.");
		return CAP_INSERT_OK;
	}

	/**
	 * The check_type function is a logical predicate function
	 * for the find_if algorithm which is used in the h323_remove_listener
	 * as well as in the h323_start_listener
	 */
	static lis_type chkLis;

	static bool
	check_lis_type(WrapListener *wl)
	{
		if (wl->listenerType == chkLis) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Start the H.323 listener
	 */
	lis_ret_val_t
	h323_start_listener(lis_type lis, char *listenAddress, int listenPort)
	{
		H323Listener					*newListener;
		lis_type_t						listenerType;
		list<WrapListener*>::iterator	cit;
		lis_ret_val_t					ret_val;

		if (end_point_exist() == NO) {
			return LIS_EP_ER;
		}

		ret_val = LIS_NSUP_ER;

		/* Check to see if a same type of listener exists in the listeners list */
		chkLis = lis;
		cit = find_if(listenerList.begin(), listenerList.end(), check_lis_type);

		if ((cit != listenerList.end())) {
			ret_val = LIS_EXIST;
			goto exit;
		}

		newListener = NULL;
		switch (lis) {
			case TCP:
			{
				PIPSocket::Address interfaceAddress(listenAddress);
				H323ListenerTCP *tcpListener = new H323ListenerTCP(*endPoint, interfaceAddress, (WORD)listenPort);
				newListener = tcpListener;
				listenerType = TCP;
				break;
			}
			case UDP:
				listenerType = UDP;
				break;
			default:
				listenerType = LIS_UNDEFINED;
				break;
		}

		/* if a new listener is created then setup stuff */
		if (newListener != NULL) {	
			if (!endPoint->StartListener(newListener)) {
				WRAPTRACEAPI(5, "Could not open H.323 listener on "
				     << ((H323ListenerTCP *) newListener));
				ret_val = LIS_FAILOP_ER;
				goto exit;
			}
			WRAPTRACEAPI(6, "Started listener " << *((H323ListenerTCP *) newListener));

			WrapListener *wl = new WrapListener(newListener, listenerType);

			/* Insert the listener to the listeners list */
			listenerList.insert(listenerList.end(), wl);
		} else {
			WRAPTRACEAPI(5, "Unsupported listener type (%d).");
			goto exit;
		}

		ret_val = LIS_START_OK;
	exit:
		return ret_val;
	}

	/**
	 * Configure the H.323 endpoint
	 */
	int
	h323_set_options(int noFastStart, int noH245Tunnelling, int noH245inSetup,
				int noSilenceSuppression, int jitterMin, int jitterMax)
	{
		WRAPTRACEAPI(5, "Setting endpoint options.");

		if (end_point_exist() == NO)
			return -1;

		WRAPTRACEAPI(5, "  FastStart option.");
		if (noFastStart)
			endPoint->DisableFastStart(TRUE);
		else
			endPoint->DisableFastStart(FALSE);
		WRAPTRACEAPI(5, "  H245Tunnelling option.");
		if (noH245Tunnelling)
			endPoint->DisableH245Tunneling(TRUE);
		else
			endPoint->DisableH245Tunneling(FALSE);
		WRAPTRACEAPI(5, "  H245InSetup option.");
		if (noH245inSetup)
			endPoint->DisableH245inSetup(TRUE);
		else
			endPoint->DisableH245inSetup(FALSE);
		/*if (noInBandDTMF) {
			endPoint->DisableDetectInBandDTMF(TRUE);
			cout << " XXXXX IN BAND DTMF DISABLED XXXX" << endl;
		} else
			endPoint->DisableDetectInBandDTMF(FALSE);*/
		WRAPTRACEAPI(5, "  Jitter option.");
		endPoint->SetAudioJitterDelay(jitterMin, jitterMax);

		Con_noFastStart = endPoint->IsFastStartDisabled();
		Con_noH245Tunnelling = endPoint->IsH245TunnelingDisabled();
		Con_noH245inSetup = endPoint->IsH245inSetupDisabled();
		/*Con_noInBandDTMF = endPoint->DetectInBandDTMFDisabled();*/
		Con_noSilenceSuppression = noSilenceSuppression;
		Con_jitterMin = jitterMin;
		Con_jitterMax = jitterMax;

		return(0);
	}

	int
	h323_set_gk(gkmode_t mode, char *gkname, char *gkpass, char **alias_tab, 
				int alias_num)
	{
		int i;

		WRAPTRACEAPI(5, "Configuring gatekeeper.");

		if (end_point_exist() == NO)
			return -1;

		/* Set the gatekeeper password */
		if (gkpass != NULL) {
			if (strlen(gkpass)) {
				PString *gkPass = new PString(gkpass);
				endPoint->SetGatekeeperPassword(*gkPass);
			}
		}

		if (gkname == NULL)
			return(0);

		rasChannel = new H323TransportUDP(*endPoint);
		PString *gkName = new PString(gkname);

		/* Set the local user name and our aliases */
		if ((alias_num <= 0)||(alias_tab == NULL)) {
			endPoint->SetLocalUserName("ASTERISK");
		} else {
			endPoint->SetLocalUserName(alias_tab[0]);
			for (i=1; i<alias_num; i++)
				endPoint->AddAliasName(alias_tab[i]);
		}

		switch (mode) {
			case GKMODE_DISABLE:
				WRAPTRACEAPI(5, "Disabling gatekeeper...");
				break;

			case GKMODE_DISCOVER:
				WRAPTRACEAPI(5, "Searching for gatekeeper...");
				if (endPoint->DiscoverGatekeeper(rasChannel))
					WRAPTRACEAPI(5, "Gatekeeper found.");
				else
					WRAPTRACEAPI(5, "Could not find a gatekeeper. "
							<< "Continuing without gatekeeper.");
				break;

			case GKMODE_NAME:
				WRAPTRACEAPI(5, "Setting gatekeeper...");
				if (endPoint->SetGatekeeper(*gkName, rasChannel))
					WRAPTRACEAPI(5, "Gatekeeper found.");
				else
					WRAPTRACEAPI(5, "Error registering with gatekeeper '" 
							<< *gkName << "'. " << "Continuing without gatekeeper.");
				break;

			case GKMODE_ID:
				WRAPTRACEAPI(5, "Locating gatekeeper...");
				if (endPoint->LocateGatekeeper(*gkName, rasChannel))
					WRAPTRACEAPI(5, "Gatekeeper found.");
				else
					WRAPTRACEAPI(5, "Error registering with gatekeeper '"
							<< *gkName << "'. " << "Continuing without gatekeeper.");
				break;

			default:
				WRAPTRACEAPI(5, "Unknown gatekeeper mode (" << mode << ").");
				break;
		}
		return(0);
	}

	/**
	 * Return the Gatekeeper name we are registered with.
	 */
	char *
	h323_get_gk(char *gk)
	{
		char *no_gk = "<No Gatekeeper>";

		WRAPTRACEAPI(5, "Returning gatekeeper.");
		
		if ((end_point_exist() == NO)||(gk == NULL)) {
			return NULL;
		}
		
		H323Gatekeeper *GK = endPoint->GetGatekeeper();
		if (GK != NULL) {
			PString gkname = GK->GetName();
			memcpy(gk, (const unsigned char*)gkname, gkname.GetLength());
			gk[gkname.GetLength()] = '\0';
		} else {
			memcpy(gk, no_gk, strlen(no_gk));
			gk[strlen(no_gk)] = '\0';
		}
		return(gk);
	}

	/** 
	 * Remove the listener with type name "lis" from the list
	 * with the current listeners. 
	 */
	lis_ret_val_t
	h323_remove_listener(lis_type lis)
	{
		list<WrapListener*>::iterator	bit, cit;
		lis_ret_val_t					ret_val;

		WRAPTRACEAPI(5, "Removing listener.");

		if (end_point_exist() == NO) {
			return LIS_EP_ER;
		}

		chkLis = lis;
		cit = find_if(bit = listenerList.begin(), listenerList.end(), check_lis_type);

		/* if the iterator points to the last element and it's type is different than
		   the requested type, then the requested type is not supported */
		if (cit == listenerList.end() || !((*cit)->listenerType == lis)) {
			ret_val = LIS_NSUP_ER;
			goto exit;
		}

		/* Remove the actual listener from the endpoint */
		endPoint->RemoveListener((*cit)->listener);
		listenerList.erase(cit);
		ret_val = LIS_REMOVE_OK;

	exit:
		return ret_val;
	}

	/**
	 * Remove all listeners from our list.
	 */
	lis_ret_val_t
	h323_removeall_listeners()
	{
		list<WrapListener*>::iterator	lit;

		WRAPTRACEAPI(5, "Removing all listeners.");

		if (end_point_exist() == NO) {
			return LIS_EP_ER;
		}

		lit = listenerList.begin();
		while (lit != listenerList.end()) {
			WRAPTRACEAPI(10, "Removing listener " << *((*lit)->listener));
			endPoint->RemoveListener((*lit)->listener);
			listenerList.erase(lit);
			lit = listenerList.begin();
		}

		return LIS_REMOVEALL_OK;
	}

	/**
	 * Remove all capabilities from our list.
	 */
	cap_ret_val_t
	h323_removeall_capabilities()
	{
		list<WrapCapability*>::iterator	cit;
		PStringArray					str_cap;

		WRAPTRACEAPI(5, "Removing all capabilities.");
		
		if (end_point_exist() == NO) {
			return CAP_EP_ER;
		}

		cit = capabilityList.begin();
		while (cit != capabilityList.end()) {
			WRAPTRACEAPI(10, "Removing capability " << ((*cit)->capability)->GetFormatName());
			str_cap[0] = ((*cit)->capability)->GetFormatName();
			endPoint->RemoveCapabilities(str_cap);
			capabilityList.erase(cit);
			cit = capabilityList.begin();
		}

		return CAP_REMOVEALL_OK;
	}

	/**
	 * Send a DTMF tone over the H323Connection with the
	 * specified token.
	 */
	void
	h323_send_tone(const char *call_token, char tone)
	{
		if (end_point_exist() == NO) {
			return;
		}

		PString token = PString(call_token);
		PString user_input = PString(tone);
		endPoint->SendUserInput(token, user_input);
	}

	/**
	 * Make a call to the remote end point.
	 */
	call_ret_val_t 
	h323_make_call(char *host, call_details_t *pcd, call_options_t call_options)
	{
		call_ret_val_t	ret_val;
		PString			token;
		unsigned int	call_reference;
		H323Capability	*outgoing_capability;

		WRAPTRACEAPI(5, "Making call.");

		if (end_point_exist() == NO) {
			return CALL_EP_ER;
		}

		PString dest = PString(host);
		outgoing_capability = h323_capability_create(endPoint, call_options.cap, 0);
		ret_val = endPoint->MakeCall(dest, token, &call_reference, 
						call_options.connectPort, outgoing_capability,
						call_options.cid);
		//WRAPTRACEAPI(6, "Token length is " << token.GetLength());
		memcpy((char *)(pcd->call_token), (const unsigned char *)token, token.GetLength());
		pcd->call_reference = call_reference;

		return ret_val;
	}

	/**
	 * Clear an established call with the remote end point.
	 */
	call_ret_val_t 
	h323_clear_call(const char *call_token)
	{
		call_ret_val_t ret_val = CALL_END_ER;

		WRAPTRACEAPI(5, "Clearing call.");

		if (end_point_exist() == NO) {
			return CALL_EP_ER;
		}

		ClearCallThread	*clearCallThread = new ClearCallThread(call_token);

		// XXX We should wait for termination and then delete
		//     the PThread object, but this causes a NULL pointer
		//     reference !!!
		//clearCallThread->WaitForTermination();
		//delete clearCallThread;

		ret_val = CALL_END_OK;
		return ret_val;
	}

	/**
	 * Answer a pending call.
	 */
	call_ret_val_t 
	h323_answer_call(const char *call_token)
	{
		call_ret_val_t ret_val;

		WRAPTRACEAPI(5, "Answering call.");

		if (end_point_exist() == NO) {
			return CALL_EP_ER;
		}

		PString token = PString(call_token);
		endPoint->AnswerCall(token);

		ret_val = CALL_ANS_OK;
		return ret_val;
	}

} /* extern "C" */

// End of file //////////////////////////////////////////////////////////////
